library flutter_graphql_client;

import 'dart:async';
import 'dart:io';

import 'package:http/http.dart';
import 'package:mime/mime.dart';
import 'package:sghi_core/communities/core/endpoints.dart';
import 'package:sghi_core/communities/core/utils.dart';
import 'package:sghi_core/flutter_graphql_client/i_flutter_graphql_client.dart';

/// [GraphQlClient] the main entry point client creation
///
/// This class is a singleton which ensues that the same object can
/// be referenced multiple times in the app
///
class GraphQlClient extends IGraphQlClient {
  GraphQlClient(String idToken, String endpoint) {
    super.idToken = idToken;
    super.endpoint = endpoint;
  }
}

// ignore: avoid_classes_with_only_static_members
/// [SimpleCall] exposes a minimal way make quickly make a call to
/// the graphql server.
///
/// By the time the call is made,its expected that [GraphQlClient]
/// has occurred. Other an authentication error will occur.
///
/// All keyword arguments are required.
///
/// if [variables] is no required, map an empty `map` instead of null
///
/// [graphClient] argument should be a valid instance of [GraphQlClient]
///
/// Example
///
/// ```dart
///
///  SimpleCall.callAPI(context, virtualCards, {},  SILAppWrapperBase.of(context).graphQLClient)
///
/// ```
class SimpleCall {
  /// [callAPI] method to call graphQL API
  static Future<dynamic> callAPI({
    required String queryString,
    required Map<String, dynamic> variables,
    required IGraphQlClient graphClient,
    bool raw = false,
    Duration? timeout,
  }) async {
    final Response result = await graphClient.query(
      queryString,
      variables,
      timeout: timeout,
    );

    // returns the raw http response without preprocessing
    if (raw) {
      return result;
    }

    final Map<String, dynamic> body = graphClient.toMap(result);

    if (graphClient.parseError(body) != null) {
      return <String, dynamic>{'error': graphClient.parseError(body)};
    } else {
      return body;
    }
  }

  /// [callRestAPI] method to call REST API
  static Future<dynamic> callRestAPI({
    required String endpoint,
    required String method,
    required IGraphQlClient graphClient,
    Map<String, dynamic>? variables,
    bool raw = false,

    /// The timeout in seconds
    Duration? timeout,
  }) async {
    final Response result = await graphClient.callRESTAPI(
      endpoint: endpoint,
      method: method,
      variables: variables,
      timeout: timeout,
    );

    // returns the raw http response without preprocessing
    if (raw) {
      return result;
    }

    final Map<String, dynamic> body = graphClient.toMap(result);

    if (graphClient.parseError(body) != null) {
      return <String, dynamic>{'error': graphClient.parseError(body)};
    } else {
      return body;
    }
  }

  static Future<Response> uploadMedia({
    required IGraphQlClient graphClient,
    required File image,
    required String mediaName,
    required String idToken,

    /// The timeout in seconds
    Duration? timeout,
    Map<String, String>? customHeaders,
  }) async {
    final String? mimeTypeOption = lookupMimeType(image.path);
    final String mimeType = convertMimeTypes(image, mimeTypeOption);

    final String fileType = mimeType;
    final String fileExtension = fileType.split('/')[1];
    final String fileName = '$mediaName.$fileExtension';
    final int fileLength = await image.length();
    final Stream<List<int>> fileStream = image.openRead();

    final Map<String, String> headers = <String, String>{
      'Authorization': 'Bearer $idToken',
      'Content-Type': fileType,
      'Content-Length': '$fileLength',
    };

    final String uri = '$uploadMediaEndpoint?filename=$fileName';

    final Response result = await graphClient.uploadMedia(
      endpoint: uri,
      uploadHeaders: customHeaders ?? headers,
      updateSink: (StreamedRequest request) {
        fileStream.listen(request.sink.add, onDone: () => request.sink.close());
      },
      timeout: timeout,
    );

    return result;
  }
}
